package de.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.ClientNotAuthenticatedException;
import de.fitko.fitconnect.api.exceptions.DecryptionException;
import de.fitko.fitconnect.api.exceptions.ValidationException;
import de.fitko.fitconnect.api.services.Subscriber;

import java.text.ParseException;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

/**
 * A fluent client for handing in a subscription for a {@link Subscriber}
 */
public class SubscriberClient {

    private SubscriberClient() {
    }

    public static RequestSubmission builder(final Subscriber subscriber, final String clientId, final String clientSecret, final String privateKey) {
        return new ClientBuilder(subscriber, clientId, clientSecret, privateKey);
    }

    public interface RequestSubmission {

        /**
         * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
         *
         * @param destinationId unique identifier for a destination
         * @param limit         number of submissions in result (max. is 500)
         * @param offset        position in the dataset
         *
         * @return list of available submissions for pickup
         */
        List<SubmissionForPickup> getAvailableSubmissions(UUID destinationId, int offset, int limit);

        /**
         * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
         *
         * @param destinationId unique identifier for a destination
         * @return list of available submissions for pickup
         */
        List<SubmissionForPickup> getAvailableSubmissions(UUID destinationId);

        /**
         * Loads a single {@link SubmissionForPickup} by id.
         * @param submissionId unique identifier of a {@link SubmissionForPickup}
         *
         * @return next step to load available {@link Metadata}, {@link Data} and {@link Attachment}s
         */
        GetData requestSubmission(UUID submissionId);
    }

    public interface GetData {

        /**
         * Gets a {@link SubmissionForPickup}´s {@link Attachment}s
         * @return list of {@link Attachment}s
         */
        List<Attachment> getAttachments();

        /**
         * Gets a {@link SubmissionForPickup}´s {@link Metadata}
         * @return the {@link Metadata}
         */
        Metadata getMetadata();

        /**
         * Gets a {@link SubmissionForPickup}´s {@link Data}
         * @return the {@link Data}
         */
        Data getData();
    }

    public static class ClientBuilder implements RequestSubmission, GetData {

        private static final int DEFAULT_SUBMISSION_LIMIT = 100;

        private final Subscriber subscriber;
        private final String clientId;
        private final String secret;
        private final String privateKey;

        public ClientBuilder(final Subscriber subscriber, final String clientId, final String clientSecret, final String privateKey) {
            this.subscriber = subscriber;
            this.clientId = clientId;
            this.secret = clientSecret;
            this.privateKey = privateKey;
            authenticate();
        }

        @Override
        public List<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId) {
            return getAvailableSubmissions(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
        }

        @Override
        public List<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
            return subscriber.pollAvailableSubmissions(destinationId, offset, limit);
        }

        @Override
        public GetData requestSubmission(final UUID submissionId) {

            final Submission submission = subscriber.getSubmission(submissionId);
            final List<Attachment> attachments = subscriber.fetchAttachments(submissionId, submission.getAttachments());
            final String encryptedData = submission.getEncryptedData();
            final String encryptedMetadata = submission.getEncryptedMetadata();

            final RSAKey decryptionKey = parseDecryptionKey(privateKey);

            final byte[] decryptedSubmissionData = subscriber.decryptStringContent(decryptionKey, encryptedData);
            final byte[] decryptedMetadata = subscriber.decryptStringContent(decryptionKey, encryptedMetadata);
            // TODO generate objects metadata and data from byte[]

            final List<ValidationResult> validationsResults = Collections.emptyList();
            validationsResults.add(subscriber.validateEventLog(submission.getCaseId()));
            validationsResults.add(subscriber.validateMetadata(getMetadata(), "schemaString"));
            validationsResults.add(subscriber.validateData(getData()));
            validationsResults.add(subscriber.validateAttachments(getAttachments()));

            final Optional<ValidationResult> validationFailed = findFailedValidation(validationsResults);

            if (validationFailed.isEmpty()) {
                subscriber.confirmValidSubmission();
                return this;
            }

            throw new ValidationException("Validation has errors, submission is not valid");
        }

        @Override
        public List<Attachment> getAttachments() {
            return null;
        }

        @Override
        public Metadata getMetadata() {
            return null;
        }

        @Override
        public Data getData() {
            return null;
        }

        private RSAKey parseDecryptionKey(final String privateKey) {
            try {
                return RSAKey.parse(privateKey);
            } catch (final ParseException e) {
                throw new DecryptionException("Key could not be parsed", e);
            }
        }

        private Optional<ValidationResult> findFailedValidation(final List<ValidationResult> validationsResults) {
            return validationsResults.stream().filter(result -> result.hasError()).findFirst();
        }

        private void authenticate() {
            final OAuthToken oAuthToken = subscriber.retrieveOAuthToken(clientId, secret);
            if (oAuthToken == null) {
                throw new ClientNotAuthenticatedException("Client is not authenticated, please authenticate first");
            }
        }

    }
}