package de.fitko.fitconnect.client.factory;

import com.typesafe.config.Config;
import com.typesafe.config.ConfigBeanFactory;
import com.typesafe.config.ConfigFactory;
import de.fitko.fitconnect.api.services.Sender;
import de.fitko.fitconnect.api.services.Subscriber;
import de.fitko.fitconnect.api.services.auth.OAuthService;
import de.fitko.fitconnect.api.services.crypto.CryptoService;
import de.fitko.fitconnect.api.services.metadata.MetadataService;
import de.fitko.fitconnect.api.services.validation.CertificateValidator;
import de.fitko.fitconnect.api.services.validation.MetadataValidator;
import de.fitko.fitconnect.client.SenderClient;
import de.fitko.fitconnect.client.SubscriberClient;
import de.fitko.fitconnect.core.SubmissionSender;
import de.fitko.fitconnect.core.SubmissionSubscriber;
import de.fitko.fitconnect.core.auth.DefaultOAuthService;
import de.fitko.fitconnect.core.crypto.JWECryptoService;
import de.fitko.fitconnect.core.http.ProxyConfig;
import de.fitko.fitconnect.core.metadata.MetadataUploadService;
import de.fitko.fitconnect.core.validation.KeyValidator;
import de.fitko.fitconnect.core.validation.MetadataSubmissionValidator;
import de.fitko.fitconnect.dependency.ApplicationConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.client.RestTemplate;

import java.io.File;

/**
 * Factory that constructs clients for {@link Sender} and {@link Subscriber}.
 */
public class ClientFactory {

    private static final Logger logger = LoggerFactory.getLogger(ClientFactory.class);

    public static final String CONFIG_FILENAME = "sdk.conf";
    public static final String CONFIG_CONTEXT = "sdk";

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} that is configured via {@link ApplicationConfig}.
     *
     * @return the sender client
     */
    public static SenderClient.WithDestination senderClient() {
        final ApplicationConfig config = loadConfig();
        final RestTemplate restTemplate = getRestTemplate(config);
        final Sender sender = getSender(config, restTemplate);
        return SenderClient.build(sender, config.getClientId(), config.getClientSecret());
    }

    /**
     * Create a new {@link SubscriberClient} that is configured via {@link ApplicationConfig}.
     *
     * @return the sender client
     */
    public static SubscriberClient.RequestSubmission subscriberClient() {
        final ApplicationConfig config = loadConfig();
        final RestTemplate restTemplate = getRestTemplate(config);
        final Subscriber subscriber = getSubscriber(config, restTemplate);
        return SubscriberClient.builder(subscriber, config.getClientId(), config.getClientSecret(), config.getPrivateKey());
    }

    private static Subscriber getSubscriber(final ApplicationConfig config, final RestTemplate restTemplate) {
        final OAuthService oAuthService = getOAuthService(config, restTemplate);
        final CryptoService cryptoService = getCryptoService();
        final MetadataValidator validator = new MetadataSubmissionValidator();
        return new SubmissionSubscriber(oAuthService, cryptoService, validator);
    }

    private static Sender getSender(final ApplicationConfig config, final RestTemplate restTemplate) {
        final OAuthService authService = getOAuthService(config, restTemplate);
        final CryptoService cryptoService = getCryptoService();
        final CertificateValidator validator = new KeyValidator();
        final MetadataService metadataService = new MetadataUploadService(restTemplate);
        return new SubmissionSender(authService, cryptoService, validator, metadataService);
    }

    private static CryptoService getCryptoService() {
        return new JWECryptoService();
    }

    private static OAuthService getOAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        return new DefaultOAuthService(restTemplate, config.getActiveEnvironment().getAuthTokenUrl());
    }

    private static RestTemplate getRestTemplate(final ApplicationConfig config) {
        final ProxyConfig proxyConfig = new ProxyConfig(config.getHttpProxyHost(), config.getHttpProxyPort());
        return proxyConfig.proxyRestTemplate();
    }

    private static ApplicationConfig loadConfig() {
        final Config configFile = ConfigFactory.parseFile(new File(CONFIG_FILENAME));
        final Config sdkConfig = ConfigFactory.load(configFile).getConfig(CONFIG_CONTEXT);
        final ApplicationConfig applicationConfig = ConfigBeanFactory.create(sdkConfig, ApplicationConfig.class);
        logger.info("Using sdk environment config {} with {}", applicationConfig.getMode(), applicationConfig);
        return applicationConfig;
    }
}




